/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include "CuttingPlanesEngine.h"
#include <cstdlib>
#include <vector>
#include <cmath>
#include <ilcplex/ilocplex.h>                    
using namespace std;

#ifdef USE_CPU_TIME_FUNCTION
#include "general.h"
#else
#include <time.h>
double getCPUTime(){
    return clock()/CLOCKS_PER_SEC; //attention to the wrap around issue, in man 3 clock
}
#endif


double lowerBound;                             //a global visible anywhere
struct cplexCheshireData{                        //From other files, one can refer to this struct
    IloEnv         env;                          //without needing to include cplex headers
    IloModel       model;
    IloNumVarArray vars;
    IloCplex       cplex ;
    IloNumArray    lb ;
    IloNumArray    ub ;
    IloRangeArray cuts;                     
    cplexCheshireData(int n):
            env(), model(env), vars(env,n,0,10000000), cplex (env),
            lb(env,n), ub (env,n), cuts(env){
        //it can assign -EPSILON to some variables >0
        //additional constraints below help avoiding this
        for(int i=0;i<n;i++){
            model.add(vars[i]>=0);
            lb[i]=0;
        }
    };

};

void CuttingPlanesEngine::removeVar(int outIdx){
    d.vars[outIdx].end();
    d.vars.remove(outIdx);
    d.lb.remove(outIdx);
    d.ub.remove(outIdx);
    n--;
}
void CuttingPlanesEngine::addVar(double objCoef, double min, double max, int swapOK){
    IloObjective obj = d.cplex.getObjective();
    d.vars.add(obj(objCoef));
    n++;
    d.lb.add(min);
    d.ub.add(max);

    d.model.add(d.vars[n-1]>=d.lb[n-1]);
    d.model.add(d.vars[n-1]<=d.ub[n-1]);
    if(swapOK){
        IloNumVar tmp = d.vars[n-1];
        d.vars[n-1] = d.vars[n-2];
        d.vars[n-2] = tmp; 
        double  tmpd= d.lb[n-1];
        d.lb[n-1]   = d.lb[n-2];
        d.lb[n-2]   = tmpd;
        tmpd= d.ub[n-1];
        d.ub[n-1]   = d.ub[n-2];
        d.ub[n-2]   = tmpd;
    }
    //delete[] primals;
    if(primals!=NULL)
        delete[] primals;
    primals = new double[n];
}
void CuttingPlanesEngine::addVar(double objCoef, double min, double max){
    return addVar(objCoef,min,max,0);
}
void CuttingPlanesEngine::init(){
    //by default there is no log output
    cplog.setstate(std::ios_base::failbit);
    ::lowerBound = INT_MIN;
    timeoutSet = -1;
    #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
    suboptimal = 0;
    #endif

    d.cplex.extract(d.model);
    d.cplex.setParam(IloCplex::Threads, THREADS);
    //For numerical precision in solving LPs
    d.cplex.setParam(IloCplex::NumericalEmphasis, true);

    #ifdef NO_CPLEX_OUTPUT
    d.cplex.setOut(d.env.getNullStream());
    d.cplex.setError(d.env.getNullStream());
    //This says when it is MIP
    d.cplex.setWarning(d.env.getNullStream());
    #endif 
    primals  = new double[n];
    noRows = 0;
    tmOnlySolve = 0;
    intVars = 0;
    intStatus = new int[n];
    totalNrCoefs      = 0;
}
int  CuttingPlanesEngine::nbIntVars(){
    return intVars;
}
void CuttingPlanesEngine::setToleranceParamToEpsilon(){
     //the first below seems an old version (<Cplex 2.6) of second below
     d.cplex.setParam(IloCplex::EpInt, EPSILON);
     #ifndef CPLEXVER
        d.cplex.setParam(IloCplex::Param::MIP::Tolerances::Integrality,EPSILON);//comment this for cplex <12.6 and maybe also use SYSTEM=x86-64_sles10_4.1 in Makefile
     #else
        //CPLEXVER passed in command line via -D. It should contain 3 digits
        #if (CPLEXVER>=126)
            d.cplex.setParam(IloCplex::Param::MIP::Tolerances::Integrality,EPSILON);//you said cplex >=12.6
        #endif
     #endif
}

void CuttingPlanesEngine::turnVarInteger(int i){
     if(intStatus[i]==1){
        cplog<<"Variable " << i <<" is already integer, can not re-turn it integer.";
        return;
     }
     cplog<<"Turning variable " << i <<" integer...";
     d.model.add(IloConversion(d.env, d.vars[i], ILOINT));
     setToleranceParamToEpsilon();
     intVars ++;
     intStatus[i] = 1;
     cplog<<"Done\n";
}
void CuttingPlanesEngine::turnAllVarsInteger(){
    if(intVars==0){
        d.model.add(IloConversion(d.env, d.vars, ILOINT));
        setToleranceParamToEpsilon();
        intVars =n ;
        for(int i=0;i<n;i++)
             intStatus[i] = 1;
        return ;
    }
    if(intVars<n){
        cplog<<"\n\nMaking all variables DISCRETE******************************************************** \n\n\n";
        //d.model.add(IloConversion(d.env, d.vars, ILOINT));
        for(int i=0;i<n;i++)
            turnVarInteger(i);
    }
}

void CuttingPlanesEngine::activateLog(){
    cplog.clear();
}
void CuttingPlanesEngine::printProgressMsg(ostream& cstream){
    cplog.rdbuf(cstream.rdbuf());
}
CuttingPlanesEngine::CuttingPlanesEngine(int nrVars, cutSeprtSimple_t cutSeprtSimple): 
                           cplog(clog.rdbuf()),//disabled by default via failbit
                           n(nrVars), lwBound(0), d(*new cplexCheshireData(n)),
                           internalCutSeprtSimple(cutSeprtSimple),
                           internalCutSeprtSolver(NULL),
                           internalCutSeprtExtended(NULL),
                           maxMoreConstr(0),
                           turnIntegerEnd(0){
    init();
};
CuttingPlanesEngine::CuttingPlanesEngine(int nrVars, cutSeprtSolver_t cutSeprtSolver): 
                           cplog(clog.rdbuf()),//disabled by default via failbit
                           n(nrVars), lwBound(0), d(*new cplexCheshireData(n)),
                           internalCutSeprtSimple(NULL),
                           internalCutSeprtSolver(cutSeprtSolver),
                           internalCutSeprtExtended(NULL),
                           maxMoreConstr(0),
                           turnIntegerEnd(0){
    init();
};
CuttingPlanesEngine::CuttingPlanesEngine(int nrVars, cutSeprtExtended_t cutSeprtExtended, int maxMoreConstrToReturn): 
                           cplog(clog.rdbuf()),//disabled by default via failbit
                           n(nrVars), lwBound(0), d(*new cplexCheshireData(n)),
                           internalCutSeprtSimple(NULL),
                           internalCutSeprtSolver(NULL),
                           internalCutSeprtExtended(cutSeprtExtended),
                           maxMoreConstr(maxMoreConstrToReturn),
                           turnIntegerEnd(0){
    init();
};

CuttingPlanesEngine::~CuttingPlanesEngine(){ 
    d.lb.end();
    d.ub.end();
    for(int i=noRows-1;i>=0;i--)
        modelDelCut(i);
    d.cuts.end();
    d.vars.end();
    d.model.end();
    d.env.end();
    delete[] primals;
    delete[] intStatus;
}
void CuttingPlanesEngine::setVarBounds(double * varLb, double * varUb){
    for (int i=0;i<n;i++){
        d.model.add(d.vars[i]>=varLb[i]);
        d.model.add(d.vars[i]<=varUb[i]);
        d.lb[i] = varLb[i];
        d.ub[i] = varUb[i];
    }
    d.vars.setBounds(d.lb,d.ub);
}
void CuttingPlanesEngine::setVarBounds(double varLb, double  varUb){
    for (int i=0;i<n;i++){
        d.lb[i] = varLb;
        d.ub[i] = varUb;
        d.model.add(d.vars[i]>=d.lb[i]);
        d.model.add(d.vars[i]<=d.ub[i]);
    }
    d.vars.setBounds(d.lb,d.ub);
}
void CuttingPlanesEngine::setVarLowerBoundsOnly(double varLb){
    for (int i=0;i<n;i++){
        d.lb[i] = varLb;
        d.ub[i] = IloInfinity;
        d.model.add(d.vars[i]>=d.lb[i]);
    }
    d.vars.setBounds(d.lb,d.ub);
}
void CuttingPlanesEngine::setObjCoefs(double * coefs){
    int i;
    IloExpr expr(d.env);
    for(i=0;i<n;i++)
        if(abs(coefs[i])>EPSILON)
            expr+=coefs[i]* d.vars[i];
    IloObjective obj(d.env, expr, IloObjective::Minimize);
    d.model.add(obj);
    expr.end();
}
void CuttingPlanesEngine::alwaysTurnIntegerInTheEnd(){
    turnIntegerEnd = 1;
}
int CuttingPlanesEngine::getNbCuts(){
    return noRows;
}
double CuttingPlanesEngine::getCutRightHand(int i){
    return d.cuts[i].getLB();
}
void CuttingPlanesEngine::modelDelCut(int i){
    d.cuts[i].end();
}
int CuttingPlanesEngine::modelAddCut(double * coefs, double rightHand){
    int i;
    IloExpr expr(d.env);
    for(i=0;i<n;i++)
        if(abs(coefs[i])>EPSILON_FOR_CUT_COEFS){
            totalNrCoefs+=1;
            expr+=coefs[i]* d.vars[i];
        }
    cplog<<"Cut Generated:";
    for(i=0;i<n;i++)
        if(abs(coefs[i])>EPSILON_FOR_CUT_COEFS)
                    cplog<<coefs[i]<<"*y["<<i<<"]+";
    cplog<<">="<<rightHand<<endl; 
    d.cuts.add(expr>=rightHand);
    noRows++;
    d.model.add(d.cuts[noRows-1]);
    expr.end();    
    return noRows-1;
}


int CuttingPlanesEngine::modelAddCut(int * coefs, double rightHand){
    int i;
    IloExpr expr(d.env);
    for(i=0;i<n;i++){
            expr+=coefs[i]* d.vars[i];
            totalNrCoefs++;
        }
        
    d.cuts.add(expr>=rightHand);
    noRows++;
    d.model.add(d.cuts[noRows-1]);
    expr.end();    
    return noRows-1;
}


double CuttingPlanesEngine::getObjVal(){
    return lwBound;
}
//set ::lowerBound = INT_MAX if error (not enough time as set by //setTimeoutSolve())
double CuttingPlanesEngine::solve(){
    double objVal=INT_MIN;
    #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
    suboptimal = 0;
    #endif
    try{
        double startTmSolve = getCPUTime();
        d.cplex.solve();
        tmOnlySolve+=(getCPUTime()-startTmSolve);
        objVal = d.cplex.getObjValue();   //Extract solution
        //cplog<<"Cut Planes objVal="<<objVal<<endl;
        //cplog<<"Status:"<<d.cplex.getCplexStatus()<<endl;
        //if(d.cplex.getCplexStatus() ==IloCplex::AbortTimeLim) cplog<<"Time limit\n";
        if(d.cplex.getCplexStatus() ==IloCplex::AbortTimeLim){
            #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
            if(timeoutSet!=-1){//-1 means disabled
            #endif
                 cerr<<"\n\n\nATTENTION: time limit "<<timeoutSet<<" secs exceeded in ConstrGener. I will report this and\n\
                       and probably stop afterwords. Did you use setTimeoutSolve()?\n\n";
                 ::lowerBound = INT_MAX;
                 lwBound    = INT_MAX;
                 return     INT_MAX;
            #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
            }
            cplog<<"\n\n\nATTENTION: macro TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS activated in Cutting Planes \n\
                       Its value is "<<TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS<<"s. The objVal is "<<objVal<<" Explanation:\n\
                       It is the number of seconds before stopping a call to solve() and take\n\
                       the current non-optimal solution. The cutting-planes can continue\n\
                       and maybe even the current non-optimal solution can be cut.\n\
                       Otherwise, I will multiply this timeout by 100 and try a full solve.\n\
                       If it fails, I'll stop. For the moment I report the primals to cutSeprt,\n\
                       I set suboptimal=1 and don't update the lower bound"<<::lowerBound<<"\n";
            objVal = INT_MAX-1;
            IloNumArray solution(d.env);
            d.cplex.getValues (d.vars,solution);
            for(int i=0;i<n;i++)
                primals[i] = solution[i];
            suboptimal = 1;
            return lwBound;//old lower bound, I don't update it when suboptimal
            #endif
        }
    }catch (IloCplex::Exception e){
        if(d.cplex.getStatus()==IloAlgorithm::InfeasibleOrUnbounded){
            cerr<<"Attention: InfeasibleOrUnbounded. Last time I had this error, it was because\n \
                   the program you served me was infeasible, it was not possible to find the \n \
                   configurations that do satisfy all set-covering constraints\n\
                   You can think of temporarily adding the 1 1 1 ... 1 columns with cost INT_MAX\n\
                   Later, you can erase it with modelDelCut(i) where i is the return value of \n\
                   the call modelAddCut for the above temporarily column. You do this when \n\
                   this column reaches a dual value of 0, via getCutCoefSol(...)\n\
                   Model saved to trycatch.lp";
            cerr<<"I do not call exit but set lwBound=INT_MAX, because this situation might be normal.  You should check this.\n"; 
            d.cplex.exportModel("trycatch.lp");
            objVal = INT_MAX;
            ::lowerBound = INT_MAX;
            lwBound = INT_MAX;
        }
    }
    if(d.cplex.getStatus()==IloAlgorithm::Infeasible){
        d.cplex.exportModel("trycatch.lp");
        cerr<<"ATTENTION on Cutting-Planes: finding infeasible solution.  Model saved to trycatch.lp. \n";
        cerr<<"I do not call exit but set lwBound=INT_MAX, because this situation might be normal.  You should check this.\n"; 
        objVal = INT_MAX;
        ::lowerBound = INT_MAX;
        lwBound = INT_MAX;

    }

    IloNumArray solution(d.env);
    d.cplex.getValues (d.vars,solution);
    for(int i=0;i<n;i++){
            primals[i] = solution[i];
            //cplog<<primals[i]<<",";
            //assert(primals[i]>=d.lb[i]);
            //assert(d.lb[i]==0);
            //use below if you don't set NumericalEmphasis true
            assert(primals[i]>=d.lb[i]-EPSILON);
            //rounding the primals to the correct values
            //if(primals[i]<d.lb[i]+EPSILON)primals[i]=d.lb[i];
    }
    //cplog<<"\n";
    
    if(objVal>lwBound){                    //Update bound
        lwBound = objVal;
        ::lowerBound = objVal;
    }
    return lwBound;
}
void CuttingPlanesEngine::getPrimals(double *ystar){
    int i;
    if(ystar!=NULL)
        for(i=0;i<n;i++)
            ystar[i] = primals[i];
}


double CuttingPlanesEngine::coefCutContrib(int coef, int idx){
    if(abs(coef)<EPSILON) //Useless when coef is integer
        return 0;
   return (double)coef * primals[idx]; 
}

double CuttingPlanesEngine::leftHandSum(int * coefs){
    double totalSum = 0;
    for (int i=0;i<n;i++)
        totalSum += coefCutContrib(coefs[i],i);
    return totalSum;
}

int CuttingPlanesEngine::violatedCut(int * coefs, double rightHand){
    double redCost = leftHandSum(coefs);
    return (redCost < rightHand - EPSILON);
}
void CuttingPlanesEngine::exportModel(char* filename){
    d.cplex.exportModel(filename);
}
void CuttingPlanesEngine::freeData(double*&a,double*&b, double**&c,int nr){
    delete[] a;
    if(internalCutSeprtExtended!=NULL){
        delete[] b;
        int i;
        for(i=0;i<nr;i++)
            delete[] c[i];
        delete[] c;
    }
}

double CuttingPlanesEngine::runSelectedCutSeprt(const int n, double*primals, double * newCut,double&newRightHand,
                                     int it, double tm, double ** newCutMore,double*newRightHandMore,
                                     int&newMore, int maxMoreConstr){
      if(internalCutSeprtExtended!=NULL)
           return internalCutSeprtExtended(n,primals,newCut, newRightHand,it,tm,
                                   newCutMore,newRightHandMore,newMore, maxMoreConstr);
      if(internalCutSeprtSolver!=NULL)
           return internalCutSeprtSolver(n,primals,newCut, newRightHand,it,tm);
      return internalCutSeprtSimple(n,primals,newCut, newRightHand);
} 
void CuttingPlanesEngine::setTimeoutSolve(double timeOut){
        d.cplex.setParam(IloCplex::ClockType,1);//timeout expressed in CPU time
        d.cplex.setParam(IloCplex::TiLim,timeOut);
        timeoutSet = timeOut;
}

int CuttingPlanesEngine::runCutPlanes(const int itMax, const double tmMax, int& it, double&tm){
        #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
        if(timeoutSet==-1){//user defined timeout has priority
            d.cplex.setParam(IloCplex::ClockType,1);
            d.cplex.setParam(IloCplex::TiLim,TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS);
        }
        #endif
  try{
       it = 0;
       tm = 0;
       double newRightHand, newRedCost,startTm = getCPUTime();
       #ifndef USE_CPU_TIME_FUNCTION
           cplog<<"I do not have a getCPUTime() function. Ignoring all time info\n";
       #endif
       double* newCut = new double[n];
   
       double** newCutMore = NULL;        //not necessarily used 
       double* newRightHandMore=NULL;     //only for cutSeprtExtended
       int newMore;
       if(internalCutSeprtExtended!=NULL){
           newCutMore = new double*[maxMoreConstr];
           newRightHandMore = new double[maxMoreConstr];
           int iii;
           for(iii=0;iii<maxMoreConstr;iii++)
                newCutMore[iii] = new double[n];
       }

       //if(!d.cplex.isDualFeasible())                        //if unbounded
       //    setVarBounds(-IloInfinity,INT_MAX/((double)n));  //Add some bounds on the variables

       solve();
       do{
           try{
           newRedCost = runSelectedCutSeprt(n,primals,newCut, newRightHand,it,tm,
                                newCutMore,newRightHandMore, newMore, maxMoreConstr);
           if(newRedCost==INT_MAX){//gap closed
                 freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
                 tm = getCPUTime() - startTm;
                 return EXIT_SUCCESS;
           }
           #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
           if(suboptimal){
                cplog<<"\n\nThe cutSeprt reports the following redCost on suboptimal solution"<<newRedCost<<endl;
                cplog<<"If this is negative, the suboptimal solution is anyway cut and I go on normally\n\n";
           }
           #endif 
           }catch (IloCplex::Exception e){
                cerr<<"\n\nConcert Exception in your cutSeprt called by the Cutting-Planes:"<<e.getMessage()<<endl;
                cerr<<"I do not call exit, because this might be normal.  You should know better, maybe catch the exception in your cutSeprt.\n\n\n"; 
                ::lowerBound = INT_MAX;
                lwBound = INT_MAX;
                freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
                tm = getCPUTime() - startTm;
                return EXIT_FAILURE;
           }
           if(newRedCost >= -EPSILON){
               #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
               if(suboptimal){
                    cplog<<"Could not cut current suboptimal solution\n";
                    cplog<<"Adding generated cuts anyways\n";
                    modelAddCut(newCut,newRightHand);
                    if(internalCutSeprtExtended!=NULL){
                         int ii;
                         for(ii=0;ii<newMore;ii++)
                             modelAddCut(newCutMore[ii], newRightHandMore[ii]);
                    }
                    cplog<<"Allowing 100 times more time, i.e., "<<100*TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS<<" seconds.\n";
                    d.cplex.setParam(IloCplex::TiLim,100*TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS);
                    solve();
                    it++;
                    if(suboptimal){
                        cerr<<"Even after allowing 100 times more time, Cutting-Planes  can not finish solve. return INFEASIBLE";
                        cerr<<"It could be possible to use CPX_PARAM_OBJLLIM to try to gain at least one\n";
                        freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
                        tm = getCPUTime() - startTm;
                        return EXIT_FAILURE;
                    }
                    if(::lowerBound==INT_MAX){//solve sets ::lowerBound=INT_MAX if infeasible
                        cerr<<"\n\n I DID try 100 times more time. Quite strange finding infeasibility here.";
                        cerr<<"\nATTENTION: THE CONSTRAINT/COL GENERATOR LP BECOME INFEASIBLE. STOP HERE\n \
                                without calling exit because this situation might be normal. Check it. \n\n\n";
                        break;
                    }
                    cplog<<"It seems I could solve it. Hope everything goes back to normal\n";
                    d.cplex.setParam(IloCplex::TiLim,TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS);
                    continue;
               }
               else{//not suboptimal, but classical case
               #endif
                 if(turnIntegerEnd){
                     if(intVars==n){           //all integer => exit
                          freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
                          tm = getCPUTime() - startTm;
                          return EXIT_SUCCESS;
                     }
                     turnAllVarsInteger();
                 }else{
                      freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
                      tm = getCPUTime() - startTm;
                      return EXIT_SUCCESS;
                 }
               #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
               }
               #endif
           }
           modelAddCut(newCut,newRightHand);
           if(internalCutSeprtExtended!=NULL){
                int ii;
                for(ii=0;ii<newMore;ii++)
                    modelAddCut(newCutMore[ii], newRightHandMore[ii]);
           }
           solve();
           if(::lowerBound==INT_MAX){//solve sets ::lowerBound=INT_MAX if infeasible
                cerr<<"\n\n\nATTENTION: THE CONSTRAINT/COL GENERATOR LP BECOME INFEASIBLE. STOP HERE\n \
                       without calling exit because this situation might be normal. Check it. \n\n\n";
                break;
           }
           //cplog<<"  -> "<<upBound<<endl;
           it ++;
           tm = getCPUTime() - startTm;
           cplog<<"ITER="<<it<<" TIME="<<tm<<" LBND="<<lwBound<<"        //message from Cutting Planes engine"<<endl;
       }while((tm<=tmMax)&& it<=itMax);
       freeData(newCut,newRightHandMore,newCutMore,maxMoreConstr);
       return EXIT_FAILURE;                                 //not success
  }
  catch (IloCplex::Exception e){
       d.cplex.exportModel("trycatch.lp");
       cerr<<"\n\nConcert Exception Cutting-Planes:"<<e.getMessage()<<endl;
       cerr<<"Model saved to trycatch.lp. I. I think the problem is infeasible. \n";
       cerr<<"I do not call exit, because this infeasibility/error might be normal.  You should check this.\n"; 
       return EXIT_FAILURE;
  }
}

int CuttingPlanesEngine::runCutPlanes(int& it, double&tm){
    return runCutPlanes(INT_MAX, INT_MAX, it, tm);
}
double CuttingPlanesEngine::getTmOnlySolve(){
    return tmOnlySolve;
}
long CuttingPlanesEngine::getTotalNrCoefs(){
    return totalNrCoefs;
}

