/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#ifndef CUTPLANESENGINE_H_INCLUDED
#define CUTPLANESENGINE_H_INCLUDED
#include<cassert>
#include<fstream>
using namespace std;

//Define below if you do have a function getCPUTime()
#define USE_CPU_TIME_FUNCTION

//The number of threads to use in cplex solver. Let 0 the default and it will choose 
//the nr of CPUs or (hyper-threads)
#define THREADS 1

//The number of seconds before stopping a call to solve() and take the current
//non-optimal solution. The cutting-planes can continue and maybe even the 
//current non-optimal solution can be cut. Otherwise, we multiply this timeout
//by ten and try a full solve. If it fails, stop.
#define TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS (n/120 +1)

//Comment or uncomment below to enable/disable terminal output
#define NO_CPLEX_OUTPUT

//this file should be stand-alone, so we define a specific epsilon
#define EPSILON 1.0e-6
//When adding new cuts, the separator might provide very small coefficients.
//Any coefficient less than below value is considered zero
#define EPSILON_FOR_CUT_COEFS 1.0e-7

//                          SEPARATOR FUNCTION (see USAGE EXAMPLE below)
//given current solution primals (dimension nrVars), find a new cut newCut and the
//associated rightHandVal. Writing a=newCut et b=rightHandVal, the added cut is:
//                               a^T x >= b
//Return value: a^T x - b, i.e., a negative value if x violates newCut, positive otherwise
//If the cut separator realizes the gap is closed, it returns INT_MAX and cutPlanes stops
//Attention: one can modify primals, but it is not recommended
typedef double (*cutSeprtSimple_t) (const int nrVars, double*primals, 
                                  double * newCut, double&newRightHand);
//Same as above, but it sends to the separator the current iteration it and time tm 
typedef double (*cutSeprtSolver_t) (const int nrVars, double*primals, 
                                 double * newCut, double&newRightHand
                                 ,int it, double tm);
//A separator that can return multiple cuts
typedef double (*cutSeprtExtended_t) (const int nrVars, double*primals, 
                                  double * newCut, double&newRightHand,
                                  int it, double tm,
                                  double **newCutMore, double*newRightHandMore,
                                  int & newMore, int maxMoreLen);

extern double lowerBound;         //global variable visible in the cut separator

class CuttingPlanesEngine{
    private:
    ostream cplog;                  //by default the log is disabled, use activateLog
    int n;                          //the total number of decision variables
    int intVars ;                   //the number of integer decision variables (<=n)
    int noRows;                     //number of rows (actually columns/configs)
    double * primals;
    double lwBound;                 //the optimum of the restricted program 
    int * intStatus;                //indicate which variables are integer
    class cplexCheshireData & d;    //d-reference (see d-pointer, compiler firewall or cheshire data)
                                    //need no cplex include outside cplex.cpp, faster compilation
                                    //At the file bottom,you find the stack-overflow post used to it
    long totalNrCoefs;              //The total number of non-zero coefficients added by the cuts
    void init();
    cutSeprtSimple_t internalCutSeprtSimple;
    cutSeprtSolver_t internalCutSeprtSolver;
    cutSeprtExtended_t internalCutSeprtExtended;
    int maxMoreConstr;
    int turnIntegerEnd;
    double timeoutSet;
    double tmOnlySolve;
    #ifdef TIMEOUT_BEFORE_GOING_SUBOPTIMAL_PRIMALS
    int suboptimal;
    #endif
    public:
    //constructs object using nrVars unbounded variables
    CuttingPlanesEngine(int nrVars, cutSeprtSimple_t pSimple);
    CuttingPlanesEngine(int nrVars, cutSeprtSolver_t pComplex);
    CuttingPlanesEngine(int nrVars, cutSeprtExtended_t pExtended, int maxMoreConstrToReturn);
    //classical destructor
    ~CuttingPlanesEngine();

    //by default, this makes it show progress (debug) message to clog
    void activateLog();
    //change the output stream (usually to cout)
    void printProgressMsg(ostream& cstream);
    //the lower and upper bound on each variable is defined by *arrays* varLB and resp varUb
    void setVarBounds(double*varLB, double*varUB);
    //the lower and upper bound on each variable is defined by doubles varLB and resp varUb
    void setVarBounds(double varLB, double varUB);
    //the lower bound on each variable is defined varLB , impose no upper bounding
    void setVarLowerBoundsOnly(double varLB);
    //Set the  objective function
    void setObjCoefs(double * coefs);
    //Ask to always turn to integer at the end
    void setTimeoutSolve(double timeout);
    void alwaysTurnIntegerInTheEnd();
    //Gets the objective value (it does not verify that the model is fully optimized)
    double getObjVal();
    //Add to the model a new cut; return value: the number (id) of the column
    int modelAddCut(double * coefs, double rightHand);
    //Add to the model with integer cut coefficients
    //Return value: the number (id) of the new cut
    int modelAddCut(int * coefs, double rightHand);
    //deletes cut of nr/id i
    void modelDelCut(int i);
    //the righthand of cut nr/id i
    double getCutRightHand(int i);
    //returns the CPU time needed only to re-optimize after each cut (excluding the separation time)
    double getTmOnlySolve();
    //returns the number of recorded cuts
    int getNbCuts();
    //returns the number of integer variables
    int nbIntVars();
    void setToleranceParamToEpsilon();//an internal to set a param
    //makes variable i integer
    void turnVarInteger(int i);
    //make all decision variables discrete, do nothing if it already the case
    void turnAllVarsInteger();
    //the contribution in the reduced cost of coef at position idx
    double coefCutContrib(int coef, int idx);
    //given newCutCoefs (coefficients of a new column), calc. \sum newCutCoefs[i]*primals[i]
    double leftHandSum(int * newCutCoefs);
    //given newCutCoefs (coefficients of a new cut) and right hand, check feasibility
    int    violatedCut(int * newCutCoefs, double rightHand);
    //solve over current columns and return the value of the objective function (restricted model)
    //set lowerBound = INT_MAX if error (not enough time as set by //setTimeoutSolve())
    double solve();
    //retrieves the variables in *yStar (if non null)
    void getPrimals(double*ystar);
    //saves the model in a file
    void exportModel(char* filename);
    //releases/free all array memory 
    void freeData(double*&a,double*&b, double**&c,int nr);

    //Solve column generator using a cutSeprt routine called 'cutSeprt' 
    //Params itMax and tmMax indicate the maximum allowed time and iterations
    //Last 2 args (filled by the routine) indicate the nr. of iterations and time required to finish
    //Returns EXIT_SUCCESS (0) in case of success or -1 otherwise
    int runCutPlanes(const int itMax, const double tmMax, int& it, double&tm);
    int runCutPlanes(int& it, double&tm);

    void removeVar(int outIdx);
    void addVar(double outIdx, double lb, double ub);
    void addVar(double outIdx, double lb, double ub, int andSwapLastTwoAfterAdd);
    long getTotalNrCoefs();
    //an internal
    double runSelectedCutSeprt(const int nrVars, double*yyy, double * newRow,double&newRightHand, int it, double tm, 
                                double ** newCutMore,double*newRightHandMore, int&newMore, int maxMoreConstr);
};
#endif

/*
///                                 //USAGE EXAMPLE
ATTENTION : undefine USE_CPU_TIME_FUNCTION unless you can provide the getCPUTime function
#include <iostream>
#include <climits>
#include <algorithm>
#include "CuttingPlanesEngine.h"                               
using namespace std;

//given current solution primals (dimension nrVars), find a new cut newCut and
//the associated rightHandVal. Writing a=newCut et b=rightHandVal, the added cut is:
//                               a^T x >= b
//Return value: a^T x - b, i.e., a negative value if x violates newCut, positive otherwise
//This cut example implements the cut family a^T x >=1, for all a with 2 non-zero values
double separator (const int nrVars, double*x, double * newCut, double&rightHandVal){ 
    int indexMin = 0;
    //cout<<"Current decision variables x";
    //for(int i=0;i<nrVars;i++)
    //    cout<<x[i]<<" ";
    //cout<<endl;
    for(int i=1;i<nrVars;i++)
        if(x[i]<x[indexMin])
            indexMin = i;
    int indexSecondMin=0;
    if(indexMin==0)
        indexSecondMin=1;
    for(int i=indexSecondMin+1;i<nrVars;i++)
        if(i!=indexMin)
            if(x[i]<x[indexSecondMin])
                indexSecondMin = i;
    //Finished computing the lowest two values of decision variables x
    for (int i=0;i<nrVars;i++)
        newCut[i]=0;
    newCut[indexMin]       = 1;    
    newCut[indexSecondMin] = 1;    
    rightHandVal           = 1;

    double violationDegree=x[indexMin]+x[indexSecondMin]-rightHandVal;
    cout<<"The separator returns the following cut:";
    cout<<"x["<<indexMin<<"]+x["<<indexSecondMin<<"]>=1\n";
    return violationDegree;
}
int main(){
    int n=30;             //number of variables
    double *b;            //objective value
    b = new double[n];
    for (int i=0;i<n;i++)
        b[i] = 10; 

    CuttingPlanesEngine cutPlanes(n,separator);      
    cutPlanes.setVarBounds(0,100);
    cutPlanes.setObjCoefs(b);
    //cutPlanes.turnVarInteger(0);  //x[0] is integer
    //cutPlanes.turnVarInteger(1);  //x[0] is integer
    double * cutExample = new double[n];
    for(int i=0;i<n;i++) cutExample[i] = 1;
    cutPlanes.modelAddCut(cutExample,n/2);
    //cutPlanes.activateLog();       //-> see messages of cutting planes
    //cutPlanes.turnAllVarsInteger();//-> the variables are discrete, as in a Benders decomposition
    
    int itersUsed;
    double timeUsed;                 //Times might be reported by problematic clock() (wrap around issue in man 3 clock), 
                                     //when USE_CPU_TIME_FUNCTION is undefined in CuttingPlanesEngine.cpp and no getCPUTime() function is known
    if(cutPlanes.runCutPlanes(itersUsed, timeUsed)==EXIT_FAILURE)
        cout<<"\n\n ATTENTION: NOT enough time or iterations to optimize to the full!";
    double finalObj = cutPlanes.getObjVal();
    cout<<"\nFinal obj val="<<finalObj<<" obtained after "<<itersUsed<<" iterations.\n";
    double*x = new double[n]; 
    cutPlanes.getPrimals(x); 
    for(int i=0;i<n;i++)
        if(x[i]>EPSILON)
            cout<<"x["<<i<<"]="<<x[i]<<"; ";
    cout<<endl; 
}
*/
	
/// d-pointers are one implementation, among many, of the pimpl pattern. It is also one of the early
/// implementations: "The name 'd-pointer' stems from Trolltech's Arnt Gulbrandsen, who first introduced
/// the technique into Qt, making it one of the first C++ GUI libraries to maintain binary compatibility
/// even between bigger release." Source
/// 
/// One advantage of using macros is the option of changing some implementation details of the pattern
/// implementation in a central place at compile time. You could for example design your macros to leave
/// you the option of switching to the fast pimpl implementation at a later time without changing tons
/// of code (hopefully you won't need this if you are using pimpl :-)). Provided that you made no
/// mistakes in your macro design/implementation...
/// 
/// However, I would personally recommend avoiding macros for your pimpl implementation as they are
/// cryptic for any newcomer to your source tree. Macros create magical dialects that are often
/// error-prone and not as meaningful as the original source code. They also come with all the problems
/// associated with the C Pre Processor; it's unaware of the underlying language.
/// 
/// Personally I like to use what I call a d-reference. Instead of a pointer, you use a reference and
/// you don't have to d-reference. 8-) It looks something like this:
/// 
/// // MyClass.h
/// 
/// class MyClass
/// {
/// public:
///     MyClass();
///     ~MyClass();
/// 
///     // implementation methods
/// 
/// private:
///     class MyClassPrivate& d;
/// };
/// 
/// // MyClass.cpp
/// 
/// struct MyClassPrivate
/// {
///     int x;
/// };
/// 
/// MyClass::MyClass()
/// : d(*new MyClassPrivate)
/// {
/// 
/// }
/// 
/// MyClass::~MyClass()
/// {
///     delete &d;
/// }
/// 
/// // In methods use d.x
/// 
/// 

