/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include <iostream>
#include <cmath>
#include <climits>
#include <algorithm>
#include <fstream>
#include <iomanip>
#include <string.h>
#include "CuttingPlanesEngine.h"    
#include "bendersSubprob.h"
#include "statManager.h"
#include "iographs.h"
using namespace std;

//#define NDEBUG
int bestUb = INT_MAX;
int iterAAA = 0;
int ** g;
int ** gEdge;                 //gives the edge id of an edge in the adjacency matrix
int n;                        //number of vertices
int m;                        //number of edges
int*ii;                       //ii[i] is the first end vertex of edge i
int*jj;                       //jj[i] is the second end vertex of edge i

int gap20It    = 0;           //the time and iteration when it reaches a ub/lb gap of 20%
double gap20Tm = 0;

int    itersCutPlanes;        //these will be filled by cutPlanes.runCutPlanes(...)
double tmCutPlanes, finalObj ;
double  totCuts, tmOnlySolve;

char* solution_method_code;
int   rndSeed;

char * texlogfilename = NULL;  //When doing several runs, all individual run output can be saved in texlogfilename
ofstream texlogfile;
ostream  logImportant(NULL);   //logImportant sends either to clog or to cout (cout used only in individual runs)
int    runId=-1;
int    failedRuns  = 0;        //Hopefully, this remains at zero. Otherwise, an error is printed.
double finalObjAllTheSame =-1;
long   totalNrCoefs=0;

statManager gap20It_stat;
statManager gap20Tm_stat;
statManager itersCutPlanes_stat, tmCutPlanes_stat;
statManager totCuts_stat, tmOnlySolve_stats;

double  yBestBound =INT_MAX ;
double* yBestEver=NULL;
int     totalSmoothedFeasible = 0;
int     totalTwoNeededForSmoothedFeasible = 0;
int     totalSmoothedInfeasible = 0;
int     totalOneEnoughForSmoothedInfeasible=0;
int     totalTwoNeededForSmoothedInfeasible = 0;

double* lastInfeasibleRay = NULL;  //useful only for useful infeasibleRayReinforceBySmoothedSeparation()
double* smoothedInfeasibleRay = NULL;
double* lastInfeasibleRaySmoothedAfter = NULL;  //useful only for useful infeasibleRayReinforceBySmoothedSeparation()
double* smoothedInfeasibleRaySmoothedAfter = NULL;
statManager totalOneEnoughForSmoothedInfeasible_stat, totalSmoothedInfeasible_stat, totalTwoNeededForSmoothedInfeasible_stat,
            totalSmoothedFeasible_stat,totalTwoNeededForSmoothedFeasible_stat, totalOneEnoughFeasible_stat;
statManager totalSmoothedBoth_stat,totalOneEnoughForSmoothedBoth_stat;
//Below convex combinations need to sum up to 1
#define CONVEX_COMB_BEST 0.5
#define CONVEX_COMB_yyy  0.5
#define INFEAS_CONVEX_COMB_LAST 0.5
#define INFEAS_CONVEX_COMB_yyy  0.5

#define CLASSICALMETH_CONVEX_COMB_LAST 0.5
#define CLASSICALMETH_CONVEX_COMB_yyy 0.5
int     totalSmoothedStdSepMethod = 0;
int     totalTwoNeededForSmoothedStdSep=0;
double* yLast=NULL;
double* ySmoothed=NULL;

//The code of the methods to be run. Each method is translated into a digit string that indicates the disabled/enabled status of different options.
//For instance, for both separation (std) and intersection (int) methods, the suffix int enables the 1-indexed digits 8 and 9. This means that we add
//p=10 random cut set constraints. Actually, digit 10 is 0 in this case, and this is how we obtain p=10; 0110000199 would mean 99 cut set constraints.
#define CMD_LINE_NAMES 10
char meth[CMD_LINE_NAMES][15] = {"std", "stdrnd"   , "int"      ,"intrnd",    "intrndband3","stdrndband3","intrndband3lp","stdrndband3lp","intrndband5lp", "stdrndband5lp"};
char code[CMD_LINE_NAMES][15] = {"011", "011000011", "211100000","211100011", "21110001103","01100001103","20110001103"  ,"00100001103"  ,"20110001105"  , "00100001105"  };
char description[CMD_LINE_NAMES][100] = {"standard separation method", "standard separation method randomized", "intersection method","intersection method randomized", "intersection method with a bandwidth of 3","standard method with a bandwidth of 3", "linear relaxation, randomized intersection method with bandwidth of 3", "linear relaxation, randomized standard method with bandwidth of 3"  ,"linear relaxation, intersection method with a bandwidth of 5"  , "linear relaxation, standard method with a bandwidth pf 5 "};
void expandMethodCodename(char* methCode){
    for(int i=0;i<CMD_LINE_NAMES;i++)
        if((!strcmp(methCode,meth[i]))){
            strcpy(methCode,code[i]);
            clog<<"Method code:"<<methCode<<endl;
        }
    if((methCode[0]<'0')||(methCode[0]>'9')){
        cerr<<"Unknown method code '"<<methCode<<"'. The method code can be one of the following:\n";
        for(int i=0;i<CMD_LINE_NAMES;i++){
            cerr<<setw(20)<<meth[i]<<" :   "<<description[i]<<endl;
        }
        exit(1);
    }
}

/*
The following digits are set in the method code, meaning the following:
digit 0  the method class (0 standard separation, 1 basic new intersection, 2 new++)
digit 1  discrete only (otherwise fractional LP + discrete at last iter)
digit 2  useMainSmoothingTechnique (feasible rays in inter method or anything in standard sep method)
digit 3  infeasibleRayReinforceBySmoothedSeparation (smoothed for infeasible rays)
digit 7  use initial random cutset constraints: random same each run, unless otherwise set by digits 8-9 below
digit 8-9 use srand(time()) when digit 7 is set >=1 (addInitialConstraints()) -> random different each time
digit                             the number indicated by digits 8-9 is the nr of random init. constraints
digit 10 consider a bandwidth defined by this digit (a value in 2-9). If digit 1 not set, solve a non-integer LP
*/

int digitIsSet(int digit){
    if(strlen(solution_method_code)<(size_t)digit+1)
        return 0;
    char charStringified [2] = {0,0};
    charStringified[0]=solution_method_code[digit];
    return atoi(charStringified);
}
int addInitialConstraints(){
    return digitIsSet(7);
}
int randomInitialCutSetConstraints(){
    return digitIsSet(8)*10+digitIsSet(9);
}
int useBandwidthDiff1(){
    return digitIsSet(10);
}
int infeasibleRayReinforceBySmoothedSeparation(){
    return digitIsSet(3);
}
int useMainSmoothingTechnique(){
    return digitIsSet(2);
}
int useDiscreteOnly(){
    return digitIsSet(1);
}
int runStandardSepMethod(){
    return(solution_method_code[0]=='0');
}
int runIntersectMethodOld(){
    return(solution_method_code[0]=='1');
}
int runIntersectMethodPlusPlus(){
    return(solution_method_code[0]=='2');
}

void yUpdateBestIfBetter(double * yyy, double tStar){
     double yNewBound=0;
     for(int i=0;i<m;i++)
          if(yyy[i]>EPSILON)
              yNewBound += yyy[i]*tStar;   //first with no rounding
     if(yNewBound>20*lowerBound)
        return;
     if(yNewBound<yBestBound){
          clog<<"    Found better yCenters to be used for smoothing, boundValue of fractional feasible="<<yNewBound<<endl;
          yBestBound = yNewBound;
          if(yBestEver == NULL)
              yBestEver = new double[m];
          for(int i=0;i<m;i++)
              yBestEver[i] = yyy[i];
     }
}

int calcUpperBound(double * yyy, double tStar){
     int ub = 0;
     //clog<<"tStarCalcUb="<<tStar<<endl;
     for(int i=0;i<m;i++){
         int tmp = (int)ceil(yyy[i]*tStar-EPSILON);
         //old version int 
         //tmp = (int)(yyy[i]*tStar-EPSILON);
         //if(tmp>0) clog<<"t="<<tmp<<","<<yyy[i]<<":"<<yyy[i]*tStar<<"; ";
         ub += tmp;
     }
     //clog<<endl;
     return ub;
}
int calcNewUbCheckBestCheckLargerThanLb(double * yyy, double tStar){
     int ub = calcUpperBound(yyy,tStar);
     if(ub<lowerBound-EPSILON){
        cerr<<"Wrong upper bound ub:"<<ub<<" and ";
        cerr<<"lb="<<lowerBound;
        cerr<<"I do not exit by pay attention.\n";
	if(runId>=0)
		cerr<<"The problematic run is runId="<<runId<<endl;
        //exit(1);
     }
     if(ub<bestUb)
       bestUb = ub; 
     return ub;
}

int swichedToPhase2(){
    return bestUb<1.1*lowerBound;
}

//returns a violation degree of yyy, ie., x^Ta-b, where x=yyy, a=newRow, b=newRightHand
double cutBendersSubprob (const int nrVars, double*yyy, double * newRow,double&newRightHand,
                int it, double tm,
                double ** newColMore,double*newRightHandMore,
                int&newMore, int maxMoreConstr){ 
    logImportant<<"-----START BENDERS SEPARATOR------\n";
    if(!runStandardSepMethod())
      if(useDiscreteOnly())                 //only in discrete/integer mode, the gap can close earlier
        if(!(bestUb>(int)ceil(lowerBound)))
            return INT_MAX;                 //a return value that signals the gap is closed
    newMore = 0;
    clog<<"Input cutBendersSubprob i(endvtx1,endvtx2):y[i]....=";
    for(int i=0;i<m;i++){
        if(yyy[i]!=0){
            clog<<i;
            clog<<"("<<ii[i]<<","<<jj[i]<<"):"<<yyy[i]<<", ";
        }
    }
    clog<<endl;
    double violationDegree = 0;
    if(!runStandardSepMethod()){
      double tStar;
      int canTestRayFeasibByCuttingSmoothedSol=
        ((runIntersectMethodPlusPlus())&&(infeasibleRayReinforceBySmoothedSeparation())&&(!swichedToPhase2()));
      if((canTestRayFeasibByCuttingSmoothedSol)&&(lastInfeasibleRay!=NULL)){
            for(int i=0;i<m;i++){
                smoothedInfeasibleRay[i] = lastInfeasibleRay[i]*INFEAS_CONVEX_COMB_LAST+ yyy[i]*INFEAS_CONVEX_COMB_yyy;
                if(smoothedInfeasibleRay[i]<EPSILON)
                    smoothedInfeasibleRay[i] = 0;
            }
            logImportant<<"        In the degenerate case, I test by separation a smoothed ray.\n";
            newCut(smoothedInfeasibleRay,newRow, newRightHand);
            double leftValue = 0;
            for(int i=0;i<m;i++)
                leftValue+=newRow[i] * yyy[i];
            int theNewCutSeparatesWholeRay=((leftValue<EPSILON)&&(newRightHand>0));
            if(!theNewCutSeparatesWholeRay){           //check if the upper bound is not anyways very bad, almost equiv à smoothed ray infeasible
                 if((leftValue>EPSILON)&&(newRightHand>EPSILON)){
                        tStar = newRightHand/leftValue;
                        if(calcUpperBound(smoothedInfeasibleRay,tStar)>bestUb)//bad upper bound, smoothed ray almost fully separated by the new cut
                            theNewCutSeparatesWholeRay = 1;
                 }
            }
            if(theNewCutSeparatesWholeRay){
                totalOneEnoughForSmoothedInfeasible++;
                logImportant<<"        One smoothed cut enough to prove the ray does not touch the polytope\n";
                clog<<"scalPord(newConstr,y)="<<leftValue<<" and rightHand="<<newRightHand<<endl;
                clog<<"Saving a one-enough constraint:";
                tStar = INFEASIBLE;
                for(int i=0;i<m;i++){
                    if(newRow[i]>0)
                        clog<<"y["<<i<<"]*"<<newRow[i]<<"+";
                }
                clog<<">="<<newRightHand<<endl;
            }else{//the generated constraint does not make the ray infeasible
                totalTwoNeededForSmoothedInfeasible++;
                clog<<"One smoothed cut NOT enough to prove the ray does not touch the polytope. Calling intersection on current non-smoothed optimal sol\n";
                clog<<"One infeasible ray smoothed cut NOT enough";
                clog<<"scalPord(newConstr,y)="<<leftValue<<" and rightHand="<<newRightHand<<endl;
                clog<<"I do Save the not-enough constraint:";
                for(int i=0;i<m;i++){
                    newColMore[newMore][i]     = newRow[i];
                    newRightHandMore[newMore] = newRightHand;
                    if(newRow[i]>0)
                        clog<<"y["<<i<<"]*"<<newRow[i]<<"+";
                }
                clog<<">="<<newRightHand<<endl;
                newMore++;
                newCutCheckRayHitsP(yyy,newRow, newRightHand, tStar);
            }
            if(tStar==INFEASIBLE)
                totalSmoothedInfeasible ++;
      }else{//Classical Intersection
            newCutCheckRayHitsP(yyy,newRow, newRightHand, tStar);
      }
      if(canTestRayFeasibByCuttingSmoothedSol){
          if(lastInfeasibleRay==NULL){
                lastInfeasibleRay = new double[m]; 
                smoothedInfeasibleRay = new double[m];
          }
          for(int ij=0;ij<m;ij++)
              lastInfeasibleRay[ij] = yyy[ij];
      }
        
      clog<<"Tested case (i), if the ray hits the polytope: tStar="<<tStar<<" (-1 means infeasible, ie. the ray missed the polytope)"<<endl;
      if(tStar!=INFEASIBLE){
            int ub;
            if(runIntersectMethodPlusPlus()&&useMainSmoothingTechnique()){
              if(yBestEver!=NULL){
                double* yCentered = new double[m];
                for(int i=0;i<m;i++){
                    yCentered[i] = yBestEver[i]*CONVEX_COMB_BEST+yyy[i]*CONVEX_COMB_yyy;
                    if(yCentered[i]<EPSILON)
                        yCentered[i] = 0;
                }
                assert(newMore<maxMoreConstr);
                newIntersectionFeasibleCase(yCentered,newColMore[newMore], newRightHandMore[newMore],tStar);
                clog<<"STAB:tStar of first smoothed constraint is:"<<tStar<<endl;
                ub = calcNewUbCheckBestCheckLargerThanLb(yCentered,tStar);
                logImportant<<"iter="<<it<<"  ub  of the first smoothed constraint when the ray touches the polytope="<<ub<<"     bestUb="<<bestUb<<"     lb "<<lowerBound<<endl;
                yUpdateBestIfBetter(yCentered,tStar);

                double newSmoothedRowValue = 0;
                for(int i=0;i<m;i++)
                    if(yyy[i]>EPSILON)
                        newSmoothedRowValue += newColMore[newMore][i]*yyy[i];
                clog<<"STAB:newSmoothedRowValueForCentered="<<newSmoothedRowValue<<endl;
                clog<<"STAB:newRightHandForCentered="<<newRightHandMore[newMore]<<endl;
                totalSmoothedFeasible++;
                if(newSmoothedRowValue<newRightHandMore[newMore]-EPSILON){
                    //we do not touch newMore => last constraint is like lost *from* newColMore
                    logImportant<<"         One smoothed constraint will be enough to cut current optimal solution (ray)\n";
                    clog<<"newMore="<<newMore<<endl;
                    newRightHand = newRightHandMore[newMore];
                    for(int i=0;i<m;i++)
                        newRow[i] = newColMore[newMore][i];
                }
                else{
                    newMore ++;                         //actual found data recorded
                    clog<<"Two constraints will be returned (smoothed one not enough)\n";
                    totalTwoNeededForSmoothedFeasible++;
                    newIntersectionFeasibleCase(yyy,newRow, newRightHand, tStar);
                    yUpdateBestIfBetter(yyy,tStar);
                    int ubHere = calcNewUbCheckBestCheckLargerThanLb(yyy,tStar);
                    ub = min(ub,ubHere);
                    clog<<"Legitimate cutBendersSubprob response tStar="<<tStar<<endl;
                    logImportant<<"iter="<<it<<"  ub of second non-smoothed constraint when the ray touches the polytope="<<ubHere<<"     bestUb="<<bestUb<<"     lb= "<<lowerBound<<endl;
                    clog<<"Found the second but surely legitimate constraint\n";
                }
                delete[] yCentered;
              }else {//checked the existence of a best ever y to smooth current y and there is none=> search classical constraint
                newIntersectionFeasibleCase(yyy,newRow, newRightHand, tStar);
                yUpdateBestIfBetter(yyy,tStar);
                ub = calcNewUbCheckBestCheckLargerThanLb(yyy,tStar);
                logImportant<<"iter="<<it<<"classical feasib case  ub="<<ub<<"     bestUb="<<bestUb<<"      lb"<<lowerBound<<endl;
              }
              clog<<"Finished search for smoothed constraints"<<endl;
            }else{//older version (solution_method_code=="1...") is never uses smoothing
                newIntersectionFeasibleCase(yyy,newRow, newRightHand, tStar);
                ub = calcNewUbCheckBestCheckLargerThanLb(yyy,tStar);
                logImportant<<"iter="<<it<<"  ub="<<ub<<"     bestUb="<<bestUb<<"      "<<lowerBound<<endl;
            }
            //still feasible case
            clog<<"plotub"<<"("<<it<<","<<ub<<")"<<endl;
            if(gap20It==0){
             if(ub<1.2*lowerBound){
                gap20It = it;
                gap20Tm = tm;
                clog<<"\n\nub at 20% from lb*******************************************************************\n\n\n"<<endl;
             }
           }
      }else{//THE RAY DOES NOT TOUCH THE FEASIBLE POLYTOPE
          logImportant<<"iter="<<it<<"  ub=INT_MAX (the ray does not touch the polytope)";
          if(bestUb<INT_MAX)
                logImportant<<"   bestUb="<<bestUb;
          else
                logImportant<<"   bestUb=INT_MAX";
          logImportant<<"     lb="<<lowerBound<<endl;
          //newCut(yyy,newRow, newRightHand);
      }
      clog<<"plotlb"<<"("<<it<<","<<lowerBound<<")"<<endl;
    }else{//standard separation method
      if(useMainSmoothingTechnique()){
           if(yLast==NULL){
                yLast = new double[m];
                ySmoothed = new double[m];
                newCut(yyy,newRow, newRightHand);
           }else{
                for(int i=0;i<m;i++){
                    ySmoothed[i] = yLast[i]*CLASSICALMETH_CONVEX_COMB_LAST+ yyy[i]*CLASSICALMETH_CONVEX_COMB_yyy;
                      if(ySmoothed[i]<EPSILON)
                          ySmoothed[i] = 0;
                }
                newCut(ySmoothed,newRow, newRightHand);
                double newSmoothedRowValue = 0;
                for(int i=0;i<m;i++)
                    if(yyy[i]>EPSILON)
                        newSmoothedRowValue += newRow[i]*yyy[i];
                totalSmoothedStdSepMethod++;
                if(newSmoothedRowValue<newRightHand-EPSILON){
                    logImportant<<"One smoothed constraint will be enough\n";
                }
                else{
                    logImportant<<"Two constraints will be returned (smoothed one not enough)\n";
                    totalTwoNeededForSmoothedStdSep++;
                    assert(newMore<maxMoreConstr);
                    for(int i=0;i<m;i++){
                        newColMore[newMore][i]     = newRow[i];
                        newRightHandMore[newMore] = newRightHand;
                    }
                    newMore++;
                    newCut(yyy,newRow, newRightHand);
                }

           }
           for(int ii=0;ii<m;ii++)
                yLast[ii] = yyy[ii]; 
        }
        else{//no smoothing asked
            newCut(yyy,newRow, newRightHand);
            clog<<"plotlb"<<"("<<it<<","<<lowerBound<<")"<<endl;
        }
      clog<<"plotlb"<<"("<<it<<","<<lowerBound<<")"<<endl;
    }

    for(int i=0;i<m;i++)
        violationDegree += newRow[i]*yyy[i];
    violationDegree-= newRightHand;
    clog<<"Generated constraint has a right hand value newRightHand="<<newRightHand<<endl;
    clog<<"Final message from cutBendersSubprob is violationDegree="<<violationDegree<<endl;
    logImportant<<"-----END BENDERS SEPARATOR------\n";
    return violationDegree;
}



void printMethCodeAndParams(ostream&localout){
    localout<<"\\\\%";
    localout<<"INTERSEC (first arg)= "<<left<<setw(11)<<solution_method_code<<" ";
    localout<<"ThreadInColGen.h:"<<THREADS<<" ";
        if(runIntersectMethodPlusPlus()){
            if(infeasibleRayReinforceBySmoothedSeparation())
                localout<<"Smoothed infeasib rays  "<<infeasibleRayReinforceBySmoothedSeparation()<<"(last*"<<INFEAS_CONVEX_COMB_LAST<<"+current*"<<INFEAS_CONVEX_COMB_yyy<<")";
            if(useMainSmoothingTechnique())
                localout<<" Smoothed feasib rays param (best*"<<CONVEX_COMB_BEST<<"+current*"<<CONVEX_COMB_yyy<<"), ";
        }
        if(runStandardSepMethod()){
            if(useMainSmoothingTechnique()){
                localout<<"Smoothed (last*"<<CLASSICALMETH_CONVEX_COMB_LAST<<"+current*"<<CLASSICALMETH_CONVEX_COMB_yyy<<"), ";
            }
        }
    if(runIntersectMethodOld())
        localout<<"Used only discrete primal vars for classical old version intersection method,";
    if(useDiscreteOnly())
        localout<<"ILP only (no LP) as asked by 2nd digit of the method code";
}
int texlogfileOpenExists(){
    if(texlogfilename==NULL)
        return 0;
    texlogfile.open(texlogfilename, std::ios::out | std::ios::app);
    if (texlogfile.fail()){
        cerr<<"Could not write to"<<texlogfilename<<endl;
        return 0;
    }
    texlogfile.exceptions(texlogfile.exceptions() | std::ios::failbit | std::ifstream::badbit);
    return 1;
}
void constructInitialConstraints(CuttingPlanesEngine &cutPlanes){
    int * connectedComponent = new int[n];
    double*cutSetConstraint = new double[m];
    double demandTotal;
    for(int i=0;i<n;i++)
        connectedComponent[i] = 1;
    int nrCutsetConstraints = 3;
    clog<<"Constraints regarding "<<runId<<endl;
    if(randomInitialCutSetConstraints()){ //if you don't give a number of constraints (digits 8-9), it is 3
        unsigned int rndSeed = time(NULL);
        clog<<"rndSeed="<<rndSeed<<endl;
        if(texlogfileOpenExists()){          //write info to file provided by -fFILENAME option, if it exists
            texlogfile<<"RNDSEED="<<rndSeed<<" and nrInitialConstraints="
                    <<randomInitialCutSetConstraints()<<" at run "<<runId
                    <<" on "<<getGraphLabelForStats()
                    <<" with m="<<m<<" and n="<<n
                    <<endl;
            texlogfile.close();
        }

        srand(rndSeed);
        nrCutsetConstraints = randomInitialCutSetConstraints();
    } 
    logImportant<<"I will add "<<nrCutsetConstraints<<"  initial random cutset constraints\n";
    for(int loop=0;loop<nrCutsetConstraints;loop++){
        for(int i=0;i<n;i++)
             connectedComponent[i] = 1;
            for(int ne=0;ne<n/10;ne++)                  //random edges
                 connectedComponent[rand()%(n-1)]=0;
        logImportant<<"Generated the "<<(1+loop)<<"th random cut set...";
        if(newCutSet(connectedComponent,cutSetConstraint,demandTotal))
             cutPlanes.modelAddCut(cutSetConstraint,demandTotal);
    }//end loop nrCutsetConstraints times

    delete[] connectedComponent;
    delete[] cutSetConstraint ;
}


int alreadyChecked = 0;
void checkPrintDegrees(ostream&localout){
    int* degrees = new int[n];
    for(int v=0;v<n;v++)
        degrees[v]=0;
    for(int e=0;e<m;e++){
        degrees[ii[e]]++;
        degrees[jj[e]]++;
    }
    localout<<getGraphLabelForStats()<<"Degrees:";
    for(int v=0;v<n;v++)
        localout<<degrees[v]<<" ";
    localout<<endl;
    delete[] degrees;
}
int checkGraphConnectivity(ostream&localout){
    if(alreadyChecked)          //do not do it twice one the same graph
        return 1;
    int* connectComp = new int[n];
    for(int i=0;i<n;i++)
        connectComp[i] = 0;
    connectComp[n-1] = 1;
    fillConnectedUpToDepth(connectComp,NULL,n-1,n);
    for(int i=0;i<n;i++)
        if(!connectComp[i])
            if(b[i]>0){
                localout<<"Vertex "<<i<<" has demand "<<b[i]<<" but it is not connected to the source";
                localout<<"\nThe instance "<<getGraphLabelForStats()<<"m="<<m<<" n="<<n<<"is infeasible. I return failure.\n";
                cerr<<"Vertex "<<i<<" has demand "<<b[i]<<" but it is not connected to the source";
                cerr<<"\nThe instance "<<getGraphLabelForStats()<<"m="<<m<<" n="<<n<<"is infeasible. I return failure.\n";
                return 0;
            }
    alreadyChecked = 1;
    delete[] connectComp;
    checkPrintDegrees(logImportant);
    return 1;//ok
}
int checkGraphConnectivityWithLog(){
    if(texlogfileOpenExists()){
        int retValue = checkGraphConnectivity(texlogfile);
        texlogfile<<endl;
        texlogfile.close();
        return retValue;
    }
    return checkGraphConnectivity(logImportant);
}

void lineBoxForLastRun(ostream&localout){
    localout<<endl<<endl;
    localout<<"                               FINAL RESULTS\n";
    double bMaxDem = -1;
    for(int i=0;i<n;i++)
        bMaxDem = max(bMaxDem,b[i]);
    if(getBandwidth()>1)    
        bMaxDem*=getBandwidth();
    localout<<"+-------------------------------------------------------------------------";
    if(!useDiscreteOnly())
        localout<<"---------";
    localout<<"+\n";
    localout<<"|method code|edges| vtx |max dem|bndWd|| OPT  |";
    if(!useDiscreteOnly())                         //lp fractional case
                localout<<" INT UB |";
    localout<<" iters |CPU tm|";
    if(useDiscreteOnly())
        localout<<"restr ILP tm|";
    else
        localout<<"restr LP  tm|";
    localout<<endl;
    localout<<"|"<<left<<setw(11)<<solution_method_code;
    localout<<"|";
    int precisionReportTm = 1;
    if(tmCutPlanes>=1)
        precisionReportTm = 2;
    if(tmCutPlanes>=10)
        precisionReportTm = 3;
    if(tmCutPlanes>=1000)
        precisionReportTm = 4;
    if(tmCutPlanes>=10000)
        precisionReportTm = 5;

    localout<<setprecision(5)<<left;
    if(finalObj==INT_MAX){
        localout<<setw(16)<<setw(5)<<m<<"|"<<setw(5)<<n<<"|"<<setw(7)<<bMaxDem<<"|"<<setw(5)<<getBandwidth()<<"||"<<setw(6)<<"Failed! "
            <<"You set a time out via cutPlanes.setTimeoutSolve()?   ";
    }else{
        localout<<setw(16)<<setw(5)<<m<<"|"<<setw(5)<<n<<"|"<<setw(7)<<bMaxDem<<"|"<<setw(5)<<getBandwidth()<<"||"<<setw(6)<<setprecision(5)<<finalObj<<"|";
        if(!useDiscreteOnly()){     //lp fractional case, used often for bandwidth>0 (hardest instances)
            if(runStandardSepMethod())
                localout<<"  ----  |";
            else
                localout<<" "<<setw(7)<<bestUb<<"|";
        }
    }
    localout<<setw(7)<<itersCutPlanes <<"|"<<setw(6)<<setprecision(precisionReportTm)<<tmCutPlanes<<"|";
    double percentOnlySolve = tmOnlySolve/tmCutPlanes*100;
    localout<<"    "<<setw(4)<<left<<setprecision(3)<<percentOnlySolve<<"%   |";
    localout<<endl;
    localout<<"+-------------------------------------------------------------------------";
    if(!useDiscreteOnly())
        localout<<"---------";
    localout<<"+\n";
    localout<<endl;
}
void latexLineForLastRun(ostream&localout){
    localout<<""<<left<<setw(11)<<solution_method_code;
    int precisionReportTm = 1;
    if(tmCutPlanes>=1)
        precisionReportTm = 2;
    if(tmCutPlanes>=10)
        precisionReportTm = 3;
    if(tmCutPlanes>=1000)
        precisionReportTm = 4;
    if(tmCutPlanes>=10000)
        precisionReportTm = 5;

    localout<<setprecision(5)<<left;
    if(finalObj==INT_MAX){
        localout<<setw(16)<<getGraphLabelForStats()<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<"Failed! "
            <<"You set a time out via cutPlanes.setTimeoutSolve()?   ";
    }else{
        localout<<setw(16)<<getGraphLabelForStats()<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<setprecision(5)<<finalObj<<"&";
        if(!useDiscreteOnly()){     //lp fractional case, used often for bandwidth>0 (hardest instances)
            if(runStandardSepMethod())
                localout<<"--      &";
            else
                localout<<"ub"<<setw(6)<<bestUb<<"&";
        }
        if(!runStandardSepMethod()){
            if(gap20It==0){ //this should not really happen, why would?
                gap20It = itersCutPlanes;
                gap20Tm = tmCutPlanes;
                localout<<"G?";
            }
            else
                localout<<"G ";
            localout<<setw(5)<<gap20It<<"&"<<setw(6)<<setprecision(precisionReportTm)<<gap20Tm<<"& ";
        }
        else
            localout<<"                ";
    }
    localout<<setw(5)<<itersCutPlanes <<"&"<<setw(6)<<setprecision(precisionReportTm)<<tmCutPlanes<<"&";
    double percentOnlySolve = tmOnlySolve/tmCutPlanes*100;
    localout<<setw(4)<<left<<setprecision(3)<<percentOnlySolve<<"\\%";
    if(runIntersectMethodPlusPlus()){
        if(infeasibleRayReinforceBySmoothedSeparation())
             localout<<"smoothed infeas:"<<totalOneEnoughForSmoothedInfeasible <<"/" <<totalSmoothedInfeasible
                                                                <<"r"<<totalTwoNeededForSmoothedInfeasible<<" ";
        if(useMainSmoothingTechnique())
           localout<<"smoothed feas:"<<totalSmoothedFeasible-totalTwoNeededForSmoothedFeasible <<"/" <<totalSmoothedFeasible<<" ";
    }
    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique())
            localout<<"SmoothedOneEnghTot:"<<totalSmoothedStdSepMethod-totalTwoNeededForSmoothedStdSep <<"/" <<totalSmoothedStdSepMethod<<",";
    }

    localout<<setprecision(5);
    localout<<"NbCuts"<<setw(5)<<totCuts
          <<" TmOnlySolve "<<setw(7)<<tmOnlySolve<<endl;
}
int oneRun(){                                   //logImportant can be clog (when multiple runs) or cout (one run)
    #ifdef NDEBUG
        logImportant<<"If you want to see more messages on the progress of the method, disable NDEBUG macro.\n\
                It could be set at compilation using option -DNEBUG or in some file included in main.cpp\n";
    #endif 
    CuttingPlanesEngine cutPlanes(m,cutBendersSubprob,10);   //maximum 10 additional constraints can be returned by subprobBenders
    if(!logImportant.fail())     //if (logImportant == cout)
        cutPlanes.activateLog();
    //cutPlanes.printProgressMsg(cout);
    //cutPlanes.alwaysTurnIntegerInTheEnd();
    //cutPlanes.activateLog();
    //cutPlanes.setVarBounds(0,10000);
    double * coefsObj;
    coefsObj = new double[m];
    for(int i=0;i<m;i++)
        coefsObj[i] = 1;
    cutPlanes.setObjCoefs(coefsObj);
    if(runIntersectMethodOld()||useDiscreteOnly())
          cutPlanes.turnAllVarsInteger();
    if(addInitialConstraints())
        constructInitialConstraints(cutPlanes);
    logImportant<<"Starting constraint generator using the Benders separator indicated by constructor"<<endl;
    if(cutPlanes.runCutPlanes(100000, 50000, itersCutPlanes, tmCutPlanes)==EXIT_FAILURE){
        cerr<<"Exiting Cutting Planes Engine with ERROR. Maybe not enough time or iterations to optimize to the full (or error? Check error messages above), or there is some unconnected sink?!\n";
	    if(runId>=0){
		    cerr<<"The problematic run is runId="<<runId
                <<" on "<<getGraphLabelForStats()
                <<" with m="<<m<<" and n="<<n<<endl;
        }
        if(texlogfileOpenExists()){
            texlogfile<<"Exiting Cutting Planes Engine with ERROR. Maybe not enough time or iterations to optimize to the full (or error? Check error messages), or there is some unconnected sink?!\n";
		    texlogfile<<"The problematic run is runId="<<runId
                <<" on "<<getGraphLabelForStats()
                <<" with m="<<m<<" and n="<<n<<endl;
            texlogfile.close();
        }
        finalObj = cutPlanes.getObjVal();
        logImportant<<"Final obj val (with timeout or error)="<<finalObj<<" obtained after "<<itersCutPlanes<<" iters of CutPlanes and "<<tmCutPlanes<<" secs.\n";
        totalNrCoefs+=cutPlanes.getTotalNrCoefs();
        return 0;
    }
    finalObj   = cutPlanes.getObjVal();
    totCuts    = cutPlanes.getNbCuts();
    tmOnlySolve= cutPlanes.getTmOnlySolve();

    //Print final information
    logImportant<<"Final obj val="<<finalObj<<" obtained after "<<itersCutPlanes<<" iters of Cutting Planes and "<<tmCutPlanes<<" secs. Final solution below under the form (edgEndVtx1,edgEndVtx2):ysol [edg]=value\n";
    double*y = new double[m]; 
    cutPlanes.getPrimals(y); 
    for(int i=0;i<m;i++){
        if(y[i]!=0){
            logImportant<<"("<<ii[i]<<","<<jj[i]<<"): ";
            logImportant<<"ysol ["<<i<<"]="<<y[i]<<"; ";
        }
    }
    logImportant<<endl;
    logImportant<<endl;

    latexLineForLastRun(logImportant);
    printMethCodeAndParams(logImportant);
    logImportant<<endl;

    if(texlogfileOpenExists()){
        latexLineForLastRun(texlogfile);
        printMethCodeAndParams(texlogfile);
        texlogfile<<endl;
        texlogfile.close();
    }
    totalNrCoefs+=cutPlanes.getTotalNrCoefs();
    if(!logImportant.fail())     //if (logImportant == cout)
        lineBoxForLastRun(logImportant);
    return 1;//ok
}
void freeTabAndInitNull(double* &tab){
    if(tab!=NULL)     
        delete[] tab; 
    tab = NULL;
}
void setGlobalsRight(){
    yBestBound =INT_MAX ;
    totalSmoothedFeasible = 0;
    totalTwoNeededForSmoothedFeasible = 0;
    totalSmoothedInfeasible = 0;
    totalOneEnoughForSmoothedInfeasible=0;
    totalTwoNeededForSmoothedInfeasible = 0;
    
    freeTabAndInitNull(yBestEver);
    freeTabAndInitNull(lastInfeasibleRay);
    freeTabAndInitNull(smoothedInfeasibleRay);
    freeTabAndInitNull(lastInfeasibleRay);
    freeTabAndInitNull(lastInfeasibleRaySmoothedAfter);
    freeTabAndInitNull(smoothedInfeasibleRaySmoothedAfter);
    //for standard separation method
    totalSmoothedStdSepMethod = 0;
    totalTwoNeededForSmoothedStdSep = 0;
    freeTabAndInitNull(ySmoothed);
    freeTabAndInitNull(yLast);
    
    bestUb = INT_MAX;
    gap20It    = 0;
    gap20Tm = 0;
}
void addToStats(){
    if(finalObjAllTheSame==-1)
        finalObjAllTheSame = finalObj;
    if(abs(finalObjAllTheSame-finalObj)>EPSILON)
        cerr<<"&STAT-ATTENTION: found finalObj="<<finalObj<<" at run "<<runId<<" but the first run found finalObjAllTheSame"<<finalObjAllTheSame<<endl;
    if(finalObj==INT_MAX){
        cerr<<"&STAT-ATTENTION: failed to finish optimization\n";
        failedRuns++;
        return ;
    }
    gap20Tm_stat.addValue(gap20Tm);
    gap20It_stat.addValue(gap20It);
    itersCutPlanes_stat.addValue(itersCutPlanes);
    tmCutPlanes_stat.addValue(tmCutPlanes);
    if(runIntersectMethodPlusPlus()){
        if(infeasibleRayReinforceBySmoothedSeparation()){
            totalOneEnoughForSmoothedInfeasible_stat.addValue(totalOneEnoughForSmoothedInfeasible);
            totalSmoothedInfeasible_stat.addValue(totalSmoothedInfeasible);
            totalTwoNeededForSmoothedInfeasible_stat.addValue(totalTwoNeededForSmoothedInfeasible);
        }
        if(useMainSmoothingTechnique()){
            totalSmoothedFeasible_stat.addValue(totalSmoothedFeasible);
            totalTwoNeededForSmoothedFeasible_stat.addValue(totalTwoNeededForSmoothedFeasible);
            totalOneEnoughFeasible_stat.addValue(totalSmoothedFeasible-totalTwoNeededForSmoothedFeasible);
        }
        if((useMainSmoothingTechnique())&&(infeasibleRayReinforceBySmoothedSeparation())){
            totalOneEnoughForSmoothedBoth_stat.addValue(totalOneEnoughForSmoothedInfeasible+
                                                    (totalSmoothedFeasible-totalTwoNeededForSmoothedFeasible));
        }
    }

    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique()){
            totalSmoothedBoth_stat.addValue(totalSmoothedStdSepMethod);
            totalOneEnoughForSmoothedBoth_stat.addValue(totalSmoothedStdSepMethod-totalTwoNeededForSmoothedStdSep);
        }
    }
    totCuts_stat.addValue(totCuts);
    tmOnlySolve_stats.addValue(tmOnlySolve);
    //we do this at oneRun in the end, please accept it: totalNrCoefs+=cutPlanes.getTotalNrCoefs();
}
void printStats(){
    if(getBandwidth()>1)
        cout<<"BND"<<getBandwidth();
    cout<<"STAT"<<itersCutPlanes_stat.getNrValues()<<" ";;
    cout<<setprecision(5)<<left;
    cout<<" &"<<setw(5)<<m<<"&"<<setw(4)<<n<<"&"<<setw(6)<<finalObjAllTheSame<<"&"<<setw(15)<<getGraphLabelForStats()<<"";
    cout<<left<<setw(10)<<solution_method_code<<" ";
    cout<<"ITERS "<<itersCutPlanes_stat.getStats()<<"TIME ";
    cout<<tmCutPlanes_stat.getStats();
    cout<<setw(2)<<setprecision(2)<<100.0*tmOnlySolve_stats.getAvg()/tmCutPlanes_stat.getAvg()<<"\\% ";
    if(runIntersectMethodPlusPlus()){
        if((useMainSmoothingTechnique())&&(infeasibleRayReinforceBySmoothedSeparation())){
             cout<<"SmoothedOneEnghTot:"<<totalOneEnoughForSmoothedBoth_stat.getStats();
        }

        cout<<"\\\\%Gap20:";
        cout<<gap20It_stat.getStats();
        cout<<gap20Tm_stat.getStats();

        if(infeasibleRayReinforceBySmoothedSeparation()){
             cout<<"STABinfeas:"<<totalOneEnoughForSmoothedInfeasible_stat.getStats()<<"/" <<
                                   totalSmoothedInfeasible_stat.getStats()
                                   <<"twice"<<totalTwoNeededForSmoothedInfeasible_stat.getStats()<<" ";
        }
        if(useMainSmoothingTechnique())
           cout<<"||STABfeas:"<<totalOneEnoughFeasible_stat.getStats()<<"/"<<totalSmoothedFeasible_stat.getStats()<<" ";
    }
    if(runStandardSepMethod()){
        if(useMainSmoothingTechnique())
            cout<<"SmoothedOneEnghTot:"<<totalOneEnoughForSmoothedBoth_stat.getStats()
                 <<"\\\\%/"<<totalSmoothedBoth_stat.getStats()<<",";
        for(int loop=0;loop<=235;loop++)    //put padding
            cout<<" ";
    }
    printMethCodeAndParams(cout);
    cout<<" Total Coefs "<<totalNrCoefs<<"";
    cout<<" |NbCuts"<<totCuts_stat.getStats()
          <<" |TmOnlySolve "<<tmOnlySolve_stats.getStats();

    if(failedRuns>0)
        cout<<"\n&STAT-ATTENTION: the following number of runs failed:"<<failedRuns;
    cout<<endl;

}

void multipleRuns(char* nrRunsStr){
    logImportant.setstate(ios_base::failbit);       //ensure no log arises in cout output of multiple runs
    int runs;
    runs = atoi(nrRunsStr);
    clog<<"We will execute the following nr of runs:"<<runs<<endl;
    for(runId = 1; runId<=runs;runId++){
        clog<<"Run "<<runId<<" on "<<getGraphLabelForStats()<<endl;
        setGlobalsRight();
        if(oneRun())
            addToStats();
    }
}
void checkAllParamsAreOk(int argc, char** argv){
    for(int i=1;i<argc;i++)
        if(argv[i][0]=='-'){                      //this is an option like -fFILENAME (no spaces)
            if((argv[i][1]!='f')&&(argv[i][1]!='d')){
                cerr<<"The only final options I accept are -fFILENAME (no spaces) or -dDEMANDMAX\n";
                exit(1);
            }
        }else{
            if(i>2)
              if(!((argv[i][0]>='0')&&(argv[i][0]<='9'))){    
                cerr<<"Argument "<<i<<" does not start with a digit or with '-'. Exit\n";
                exit(1);
              }
        }
}
int main(int argc, char** argv){
    #ifdef NDEBUG
    //in running=NDEBUG mode, you can use cout for printing
    //don't disable cerr, it could be useful
    //cerr.setstate(ios_base::failbit);
    clog.setstate(ios_base::failbit);
    #endif
    checkAllParamsAreOk(argc,argv);
    if((argv[argc-1][0]=='-')&&(argv[argc-1][1]=='d')){
        bMaxDemandFromCmdLine=atoi(argv[argc-1]+2);
        cout<<"Setting a maximum demand of "<<bMaxDemandFromCmdLine<<" overriding other settings\n";
        argc--;
    }
    if((argv[argc-1][0]=='-')&&(argv[argc-1][1]=='f')){    
        texlogfilename = new char[strlen(argv[argc-1])];
        strcpy(texlogfilename,argv[argc-1]+2);
        clog<<"Writing tex results of each run in "<<texlogfilename<<endl;
        argc--;
    }
    if((argv[argc-1][0]=='-')&&(argv[argc-1][1]=='d')){
        bMaxDemandFromCmdLine=atoi(argv[argc-1]+2);
        clog<<"Setting a maximum demand of "<<bMaxDemandFromCmdLine<<" overriding other settings\n";
        argc--;
    }
    if(argc==1){
        cerr<<"\nMAIN USAGE: ./main method instanceFile [runs] \n";
        cerr<<"                   - method can be : \"std\", \"int\", \"stdrnd\", \"intrnd\", (rnd indicating randomized run), see below. \n";
        cerr<<"                   - if you do not provide a number of [runs], the program prints the progress messages of the sub-problems and the Cutting Planes.\n";
        cerr<<"                   - type .\\main help for more details.\n";
        cerr<<"The method field can be one of the following:\n";
        for(int i=0;i<CMD_LINE_NAMES;i++){
            cerr<<setw(15)<<left<<meth[i]<<description[i]<<endl;
        }
        return 1;
    }
    if((argc<3)||(argc>6)){
        cerr<<endl;
        cerr<<"Usage1: ./main method(\"std\", \"int\", \"stdrnd\", \"intrnd\") instanceFile [runs] \n";
        cerr<<"Usage2: ./main method(\"std\", \"int\", \"stdrnd\", \"intrnd\") m(edges) n(vtx) rndseed [runs] \n";
        cerr<<"                If you provide an optional number of runs [runs], the program will only print the final statistics and nothing else\n";
        cerr<<"                If you do not provide a number of runs, it will print the progress messages of the sub-problems and the Cutting Planes \n";
        cerr<<"The method field can be one of the following:\n";
        for(int i=0;i<CMD_LINE_NAMES;i++){
            cerr<<setw(15)<<left<<meth[i]<<description[i]<<endl;
        }
        return 1;
    }
    solution_method_code = new char[50];
    strncpy(solution_method_code,argv[1],45);
    expandMethodCodename(solution_method_code);
    setBandwidth(1);
    if(useBandwidthDiff1())
        setBandwidth(useBandwidthDiff1());

    if(argc>=5){//usage 1
        m       = atoi(argv[2]);
        n       = atoi(argv[3]);
        rndSeed      = atoi(argv[4]);

        genRndGraphAndDemandsOrLoadFromAssocFileWhenItExists(rndSeed);
        if(argc==5){
            if(!checkGraphConnectivity(cout))
                exit(1);
            clog<<"One standard run:\n";
            logImportant.rdbuf(cout.rdbuf());
            oneRun();
            return 0;
        }
        if(argc==6){
            if(!checkGraphConnectivityWithLog())
                exit(1);
            multipleRuns(argv[5]);
        }
    }
    if(argc<5){        //It enters here if it is 3 or 4
        if(bMaxDemandFromCmdLine>=0){
            cerr<<"You asked me to read the graph from "<<argv[2]<<"but you provided a maximum demand value "
                <<"using -dDEMANDMAX option. If I read it from file, I can not set the maximum demand this way. "
                <<"Please remove -dDEMANDMAX. Exit now";
            exit(1);
        }
        loadGraphFromFile(argv[2]);
        if(argc==3){
                if(!checkGraphConnectivity(cout))
                    exit(1);
                clog<<"One standard run on "<<argv[2]<<":\n";
                logImportant.rdbuf(cout.rdbuf());
                oneRun();
                return 0;
        }
        if(argc==4){
            if(!checkGraphConnectivityWithLog())
                exit(1);
            multipleRuns(argv[3]);
        }
    }

    #ifdef NDEBUG
    clog.clear();
    //cerr.clear();
    #endif
    printStats();
}
